import elasticsearch.exceptions
from elasticsearch.helpers import scan
from elasticsearch_dsl import Search, Q
from zope.interface import alsoProvides
from uuid import UUID
from .interfaces import (
    ELASTIC_SEARCH,
    INDEXER_QUEUE_MIRROR,
    INDEXER,
    ICachedItem,
)
from .indexer_utils import get_namespaced_index, namespace_index_from_health, find_uuids_for_indexing
from .create_mapping import SEARCH_MAX
from ..storage import register_storage
from ..util import CachedField
from dcicutils import es_utils, ff_utils
import structlog

log = structlog.getLogger(__name__)


def includeme(config):
    registry = config.registry
    # IMPORTANT: update read storage on PickStorage created in storage.py
    # Without `register_storage` call, cannot read from ES in the portal
    read_storage = ElasticSearchStorage(registry)
    register_storage(registry, read_override=read_storage)


class CachedModel(object):
    """
    Model used by resources returned from ElasticSearchStorage.
    Leverages cached_views.ICachedItem to create different item views
    that use this model.
    Analogous to storage.Resource, the model used for database resources
    """
    used_datastore = 'elasticsearch'

    def __init__(self, source):
        """
        Takes a dictionary document `source`
        """
        self.source = source

    @property
    def item_type(self):
        return self.source['item_type']

    @property
    def properties(self):
        return self.source['properties']

    @property
    def propsheets(self):
        return self.source['propsheets']

    @property
    def unique_keys(self):
        return self.source['unique_keys']

    @property
    def links(self):
        return self.source['links']

    @property
    def uuid(self):
        """
        Return UUID object to be consistent with Resource.uuid
        """
        return UUID(self.source['uuid'])

    @property
    def sid(self):
        return self.source['sid']

    @property
    def max_sid(self):
        return self.source['max_sid']

    def used_for(self, item):
        alsoProvides(item, ICachedItem)


class ElasticSearchStorage(object):
    """
    Storage class used to interface with Elasticsearch.
    Corresponds to storage.PickStorage.read and is analagous to
    storage.RDBStorage, which is used when working with DB models
    """
    def __init__(self, registry):
        self.registry = registry
        self.es = registry[ELASTIC_SEARCH]
        self.index = get_namespaced_index(registry, '*')
        self.mirror = self.registry.settings.get('mirror_health', None) is not None
        self.mirror_client = None
        # XXX: cache elastic search mappings here subject to a TTL
        # Use this field in search so you don't have to get mappings on every search
        self.mappings = CachedField('mappings',
                                    lambda: self.es.indices.get_mapping(index=self.index))

    def _one(self, search):
        # execute search and return a model if there is one hit
        hits = search.execute()
        if len(hits) != 1:
            return None
        model = CachedModel(hits[0].to_dict())
        return model

    @staticmethod
    def find_linking_property(our_dict, value_to_find):
        """
        Helper function used in ElasticSearchStorage.find_uuids_linked_to_item
        """
        def find_it(d, parent_key=None):
            if isinstance(d, list):
                for idx, v in enumerate(d):
                    if isinstance(v, dict) or isinstance(v, list):
                        found = find_it(v, parent_key)
                        if found:
                            return (parent_key if parent_key else '') + '[' + str(idx) + '].' + found
                    elif v == value_to_find:
                        return '[' + str(idx) + ']'
            elif isinstance(d, dict):
                for k, v in d.items():
                    if isinstance(v, dict) or isinstance(v, list):
                        found = find_it(v, k)
                        if found:
                            return found
                    elif v == value_to_find:
                        return k
            return None

        return find_it(our_dict)

    def get_by_uuid(self, uuid):
        """
        This calls a search, and index/doc_type does not need to be provided
        Returns a CachedModel built with the es hit, or None if it is not found

        Args:
            uuid (str): uuid of the item to find

        Returns:
            CachedModel of the hit or None
        """
        search = Search(using=self.es, index=self.index)
        id_query = Q('ids', values=[str(uuid)])
        search = search.query(id_query)
        return self._one(search)

    def get_by_uuid_direct(self, uuid, item_type):
        """
        See if a document exists under the index/doc_type given by item_type.
        self.es.get calls a GET request, which will refresh the given
        document (and all other docs) in realtime, if necessary. Explicitly
        ignore 404 responses from elasticsearch to reduce logging.

        NOTE: this function DOES NOT use CachedModel, as it is used for direct
              querying of ES items during indexing only. It might be performant
              to set the CachedModel from these results.
              Right now, I'm not doing it because get_by_uuid_direct
              is possibly called very often during indexing.

        Args:
            uuid (str): uuid of the item to GET
            item_type (str): item_type of the item to GET

        Returns:
            dict: the Elasticsearch document if found, else None
        """
        index_name = get_namespaced_index(self.registry, item_type)
        try:
            res = self.es.get(index=index_name, doc_type=item_type, id=uuid,
                              _source=True, realtime=True, ignore=404)
        except elasticsearch.exceptions.NotFoundError:
            res = None
        return res

    def get_by_json(self, key, value, item_type, default=None):
        """
        Perform a search with an given key and value.
        Returns CachedModel if found, otherwise None
        """
        # find the term with the specific type
        term = 'embedded.' + key + '.raw'
        index = get_namespaced_index(self.registry, item_type)
        search = Search(using=self.es, index=index)
        search = search.filter('term', **{term: value})
        search = search.filter('type', value=item_type)
        return self._one(search)

    def get_by_unique_key(self, unique_key, name, item_type=None):
        """
        Perform a search against unique keys with given unique_key (field) and
        name (value).
        Returns CachedModel if found, otherwise None
        """
        term = 'unique_keys.' + unique_key
        index = self.index
        if item_type:
            index = get_namespaced_index(self.registry, item_type)
        # had to use ** kw notation because of variable in field name
        search = Search(using=self.es, index=index)
        search = search.filter('term', **{term: name})
        search = search.extra(version=True)
        return self._one(search)

    def get_rev_links(self, model, rel, *item_types):
        """
        Get the reverse links given a item model and a rev link name (rel), as
        well as the applicable item types of the reverse links.
        Perform a search and return a list of uuids of the resulting documents
        """
        search = Search(using=self.es, index=self.index)
        search = search.extra(size=SEARCH_MAX)
        # rel links use '~' instead of '.' due to ES field restraints
        proc_rel = rel.replace('.', '~')
        # had to use ** kw notation because of variable in field name
        search = search.filter('term', **{'links.' + proc_rel: str(model.uuid)})
        if item_types:
            search = search.filter('terms', item_type=item_types)
        hits = search.execute()
        return [hit.to_dict().get('uuid', hit.to_dict().get('_id')) for hit in hits]

    def get_sids_by_uuids(self, rids):
        """
        Currently not implemented for ES. Just return an empty dict

        Args:
            rids (list): list of string rids (uuids)

        Returns:
            dict keyed by rid with integer sid values
        """
        return {}

    def get_max_sid(self):
        """
        Currently not implemented for ES. Just return None
        """
        return None

    @staticmethod
    def _delete_from_es(es, rid, index_name, item_type):
        """
        Helper method that deletes an item from Elasticsearch, returning a boolean success value.
        Success means the given rid did not exist or no longer exists in ES.

        :param es: es client to use
        :param rid: resource id to purge
        :param index_name: index_name to purge rid from
        :param item_type: type of rid
        :returns: True if rid does not exist in ES, False if it does (AT THIS TIME)
        """
        try:
            es.delete(id=rid, index=index_name, doc_type=item_type)
        except elasticsearch.exceptions.NotFoundError:
            # Case: Not yet indexed
            log.error('PURGE: Could not find %s in ElasticSearch. Continuing.' % rid)
        except Exception as exc:
            log.error('PURGE: Cannot delete %s in ElasticSearch. Error: %s. NOT proceeding with deletion.'
                      % (item_type, str(exc)))
            return False  # only return False here as this is the only scenario where we want to STOP the purge process
        else:
            log.info('PURGE: successfully deleted %s in ElasticSearch' % rid)
        return True

    def _get_cached_mirror_health(self, mirror_env):

        cached_mirror_health = self.registry.settings['mirror_health']
        if 'error' not in cached_mirror_health:
            return cached_mirror_health

        mirror_env = self.registry.settings['mirror.env.name']
        mirror_health_now = ff_utils.get_health_page(ff_env=mirror_env)
        if 'error' not in mirror_health_now:
            return mirror_health_now

        raise RuntimeError('PURGE: Could not resolve mirror health on retry with error: %s' % mirror_health_now)

    def _assure_mirror_client(self, es_mirror_server_and_port):
        if not self.mirror:
            raise RuntimeError("Attempt to call ._assure_mirror_client() when there is no self.mirror.")

        use_aws_auth = self.registry.settings.get('elasticsearch.aws_auth')

        # make sure use_aws_auth is bool
        if not isinstance(use_aws_auth, bool):
            use_aws_auth = True if use_aws_auth == 'true' else False

        if not self.mirror_client:  # only recompute if we've never done this before
            self.mirror_client = es_utils.create_es_client(es_mirror_server_and_port, use_aws_auth=use_aws_auth)

    def _purge_uuid_from_mirror_es(self, rid, item_type):
        """
        Helper method for purge_uuid that purges rid from the mirror on ESStorage

        :param rid: resource id to purge
        :param item_type: type of rid
        :returns: result of 'delete_from_es'
        :raises: RuntimeError if we are unable to get the mirror health page
        """

        if not self.mirror:
            raise RuntimeError("Attempt to call ._purge_uuid_from_mirror_es() when there is no self.mirror.")

        mirror_env = self.registry.settings['mirror.env.name']
        log.info('PURGE: attempting to purge %s from mirror storage %s' % (rid, mirror_env))

        try:
            mirror_health = self._get_cached_mirror_health(mirror_env)
        except RuntimeError:
            log.error("PURGE: Tried to purge %s from mirror storage but couldn't get health page. Is staging up?" % rid)
            raise

        self._assure_mirror_client(es_mirror_server_and_port=mirror_health['elasticsearch'])

        mirror_index_name = namespace_index_from_health(health=mirror_health, index=item_type)
        return self._delete_from_es(es=self.mirror_client, rid=rid, index_name=mirror_index_name, item_type=item_type)

    def _purge_uuid_from_primary_es(self, rid, item_type, max_sid=None):

        index_name = get_namespaced_index(config=self.registry, index=item_type)
        if not self._delete_from_es(es=self.es, rid=rid, index_name=index_name, item_type=item_type):
            return False

        # queue related items for reindexing if deletion was successful
        self.registry[INDEXER].find_and_queue_secondary_items({rid}, set(), sid=max_sid)

        return True

    def purge_uuid(self, rid, item_type, max_sid=None):
        """
        Purge a uuid from the write storage (Elasticsearch)
        If the indexer has an ElasticSearch mirror environment, also attempt to remove the uuid from that mirror.

        Returns True if the deletion was successful or the item was not present, and False otherwise.
        """
        if not self._purge_uuid_from_primary_es(rid=rid, item_type=item_type, max_sid=max_sid):
            return False

        # if configured, delete the item from the mirrored ES as well.
        # If that is successful, we were successful, otherwise not.
        if self.mirror:
            return self._purge_uuid_from_mirror_es(rid=rid, item_type=item_type)
        else:
            # We don't usually need over and over in logs for cgap, where it's normal for there to be no mirror.
            log.debug('PURGE: Did not find a mirror env. Continuing.')
            return True

    def __iter__(self, *item_types):
        """
        Return a generator that yields string uuids of all documents matching
        given item types (through kwargs). Use all item types if none provided
        """
        query = {'query': {
            'bool': {
                'filter': {'terms': {'item_type': item_types}} if item_types else {'match_all': {}}
            }
        }}
        for hit in scan(self.es, index=self.index, query=query):
            yield hit.get('uuid', hit.get('_id'))

    def __len__(self, *item_types):
        """
        Return an integer count of the number of documents resulting for a
        search across all given item types (through kwargs). Use all item types
        if none provided
        """
        query = {'query': {
            'bool': {
                'filter': {'terms': {'item_type': item_types}} if item_types else {'match_all': {}}
            }
        }}
        result = self.es.count(index=self.index, body=query)
        return result['count']

    def find_uuids_linked_to_item(self, rid):
        """
        Given a resource id (uuid), find all items in ES that linkTo that item.
        Returns some extra information about the fields/links that are present
        """
        linked_info = []
        # we only care about linkTos the item and not reverse links here
        # we also do not care about invalidation scope
        uuids_linking_to_item, _ = find_uuids_for_indexing(self.registry, set([rid]))
        # remove the item itself from the list
        uuids_linking_to_item = uuids_linking_to_item - set([rid])
        if 0 < len(uuids_linking_to_item) < 1000:  # more than 1000 can trigger 504
            # Return list of { '@id', 'display_title', 'uuid' } in 'comment'
            # property of HTTPException response to assist with any manual unlinking.
            for linking_uuid in uuids_linking_to_item:
                linking_dict = self.get_by_uuid(linking_uuid).source.get('embedded')
                linking_property = self.find_linking_property(linking_dict, rid)
                linked_info.append({
                    '@id' : linking_dict.get('@id', linking_dict['uuid']),
                    'display_title' : linking_dict.get('display_title', linking_dict['uuid']),
                    'uuid' : linking_uuid,
                    'field' : linking_property or "Not Embedded"
                })
        return linked_info

    def update(self, model, properties, sheets, unique_keys, links):
        """
        Update the ES contents for an ElasticSearch-based item.
        `model` is database model (storage.Resource), which is needed for
        getting correct sid/max_sid.
        This function gets existing ES document for the given item and updates
        the fields contained in `document` below. If there is no existing item,
        create an ES document with the minimal information.
        This will create a discrepancy between the
        new attributes and the indexed ones, like `properties` vs `embedded`.
        However, this is consistent with the database approach of indexing
        and the document will be updated after it is indexed
        """
        # links in ES use "~" instead of "."
        es_links = {}
        for key, val in links.items():
            es_links['~'.join(key.split('.'))] = val

        document = {
            'item_type': model.item_type,
            'uuid': str(model.uuid),
            'max_sid': model.max_sid,
            'sid': model.sid,
            'properties': properties,
            'propsheets': {},
            'unique_keys': unique_keys,
            'links': es_links
        }
        if sheets is not None:
            document['propsheets'] = sheets

        # get existing document if it exists
        existing_doc = self.get_by_uuid(document['uuid'])
        if existing_doc is not None:
            existing_doc.source.update(document)
            document = existing_doc.source

        index_name = get_namespaced_index(self.registry, document['item_type'])
        # use `refresh='waitfor'` so that the ES model is immediately available
        self.es.index(
            index=index_name, doc_type=document['item_type'], body=document,
            id=document['uuid'], version=document['sid'],
            version_type='external_gte', refresh='wait_for'
        )



import pytest

from dcicutils.qa_utils import ignored, notice_pytest_fixtures
from re import findall
from ..interfaces import TYPES
from ..util import add_default_embeds, crawl_schemas_by_embeds
from .test_views import PARAMETERIZED_NAMES


targets = [
    {'name': 'one', 'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f'},
    {'name': 'two', 'uuid': 'd6784f5e-48a1-4b40-9b11-c8aefb6e1377'},
]

sources = [
    {
        'name': 'A',
        'target': '775795d3-4410-4114-836b-8eeecf1d0c2f',
        'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd',
        'status': 'current',
    },
    {
        'name': 'B',
        'target': 'd6784f5e-48a1-4b40-9b11-c8aefb6e1377',
        'uuid': '1e152917-c5fd-4aec-b74f-b0533d0cc55c',
        'status': 'deleted',
    },
]


# Convert names to snake is needed so that
# we can parameterize on the embeds
def convert_names_to_snake():
    res = []
    for name in PARAMETERIZED_NAMES:
        caps = findall('[A-Z][^A-Z]*', name)
        snake = '_'.join(map(str.lower, caps))
        res.append(snake)
    return res


SNAKE_NAMES = convert_names_to_snake()


@pytest.fixture
def content(testapp):
    url = '/testing-link-targets-sno/'
    for item in targets:
        testapp.post_json(url, item, status=201)

    url = '/testing-link-sources-sno/'
    for item in sources:
        testapp.post_json(url, item, status=201)


@pytest.mark.parametrize('item_type', [name for name in SNAKE_NAMES if name != 'testing_server_default'])
def test_add_default_embeds(registry, item_type):
    """
    Ensure default embedding matches the schema for each object
    """
    notice_pytest_fixtures(registry)
    type_info = registry[TYPES].by_item_type[item_type]
    schema = type_info.schema
    embeds = add_default_embeds(item_type, registry[TYPES], type_info.embedded_list, schema)
    principals_allowed_included_in_default_embeds = False
    for embed in embeds:
        split_embed = embed.strip().split('.')
        if 'principals_allowed' in split_embed:
            principals_allowed_included_in_default_embeds = True
        error, _ = crawl_schemas_by_embeds(item_type, registry[TYPES], split_embed, schema['properties'])
        assert error is None

    assert principals_allowed_included_in_default_embeds


@pytest.mark.parametrize('item_type', SNAKE_NAMES)
def test_manual_embeds(registry, item_type):
    """
    Ensure manual embedding in the types files are valid
    """
    notice_pytest_fixtures(registry)
    type_info = registry[TYPES].by_item_type[item_type]
    schema = type_info.schema
    embeds = type_info.embedded_list
    for embed in embeds:
        split_embed = embed.strip().split('.')
        error, _ = crawl_schemas_by_embeds(item_type, registry[TYPES], split_embed, schema['properties'])
        assert error is None


def test_linked_uuids_unset(content, dummy_request, threadlocals):
    notice_pytest_fixtures(content, dummy_request, threadlocals)
    # without setting _indexing_view = True on the request,
    # _linked_uuids not tracked and _sid_cache not populated in resource.py
    dummy_request.embed('/testing-link-sources-sno/', sources[0]['uuid'], '@@object')
    assert dummy_request._linked_uuids == set()
    assert dummy_request._sid_cache == {}


def test_linked_uuids_object(content, dummy_request, threadlocals):
    notice_pytest_fixtures(content, dummy_request, threadlocals)
    # needed to track _linked_uuids
    dummy_request._indexing_view = True
    dummy_request.embed('/testing-link-sources-sno/', sources[0]['uuid'], '@@object')
    assert dummy_request._linked_uuids == {('16157204-8c8f-4672-a1a4-14f4b8021fcd', 'TestingLinkSourceSno')}
    assert dummy_request._rev_linked_uuids_by_item == {}


def test_linked_uuids_embedded(content, dummy_request, threadlocals):
    notice_pytest_fixtures(content, dummy_request, threadlocals)
    # needed to track _linked_uuids
    dummy_request._indexing_view = True
    dummy_request.embed('/testing-link-sources-sno/', sources[0]['uuid'], '@@embedded')
    assert dummy_request._linked_uuids == {
        ('16157204-8c8f-4672-a1a4-14f4b8021fcd', 'TestingLinkSourceSno'),
        ('775795d3-4410-4114-836b-8eeecf1d0c2f', 'TestingLinkTargetSno')
    }
    # _rev_linked_uuids_by_item is in form {target uuid: set(source uuid)}
    assert dummy_request._rev_linked_uuids_by_item == {
        '775795d3-4410-4114-836b-8eeecf1d0c2f': {'reverse': ['16157204-8c8f-4672-a1a4-14f4b8021fcd']}
    }


def test_linked_uuids_page(content, dummy_request, threadlocals):
    notice_pytest_fixtures(content, dummy_request, threadlocals)
    # needed to track _linked_uuids
    dummy_request._indexing_view = True
    dummy_request.embed('/testing-link-sources-sno/', sources[0]['uuid'], '@@page')
    assert dummy_request._linked_uuids == {
         ('16157204-8c8f-4672-a1a4-14f4b8021fcd', 'TestingLinkSourceSno'),
         ('775795d3-4410-4114-836b-8eeecf1d0c2f', 'TestingLinkTargetSno')
    }
    assert dummy_request._rev_linked_uuids_by_item == {
        '775795d3-4410-4114-836b-8eeecf1d0c2f': {'reverse': ['16157204-8c8f-4672-a1a4-14f4b8021fcd']}
    }


def test_linked_uuids_expand_target(content, dummy_request, threadlocals):
    notice_pytest_fixtures(content, dummy_request, threadlocals)
    # needed to track _linked_uuids
    dummy_request._indexing_view = True
    dummy_request.embed('/testing-link-sources-sno/', sources[0]['uuid'], '@@expand?expand=target')
    # expanding does not add to the embedded_list
    assert dummy_request._linked_uuids == {
         ('16157204-8c8f-4672-a1a4-14f4b8021fcd', 'TestingLinkSourceSno'),
         ('775795d3-4410-4114-836b-8eeecf1d0c2f', 'TestingLinkTargetSno')
    }
    assert dummy_request._rev_linked_uuids_by_item == {
        '775795d3-4410-4114-836b-8eeecf1d0c2f': {'reverse': ['16157204-8c8f-4672-a1a4-14f4b8021fcd']}
    }


def test_linked_uuids_index_data(content, dummy_request, threadlocals):
    notice_pytest_fixtures(content, dummy_request, threadlocals)
    # this is the main view use to create data model for indexing
    # automatically sets request._indexing_view and will populate
    # a number of different attributes on the request
    res = dummy_request.embed('/testing-link-sources-sno/', sources[0]['uuid'], '@@index-data', as_user='INDEXER')
    # Since the embedded view is run last, these values correspond to that view
    assert dummy_request._linked_uuids == {(sources[0]['uuid'], 'TestingLinkSourceSno'),
                                           (targets[0]['uuid'], 'TestingLinkTargetSno')}
    assert dummy_request._rev_linked_uuids_by_item == {targets[0]['uuid']: {'reverse': [sources[0]['uuid']]}}
    # Confirm all items in the _sid_cache are up-to-date
    for rid, _ in dummy_request._linked_uuids:
        found_sid = dummy_request.registry['storage'].write.get_by_uuid(rid).sid
        assert dummy_request._sid_cache.get(rid) == found_sid

    # embedded view linked uuids are unchanged
    assert res['rev_link_names'] == {}
    assert res['rev_linked_to_me'] == [targets[0]['uuid']]
    # object view linked uuids are contained within the embedded linked uuids
    assert set((linked['uuid'], linked['item_type']) for linked in res['linked_uuids_object']) <= dummy_request._linked_uuids

    # now test the target. this will reset all attributes on dummy_request
    res2 = dummy_request.embed('/testing-link-targets-sno/', targets[0]['uuid'], '@@index-data', as_user='INDEXER')
    assert dummy_request._linked_uuids == {(sources[0]['uuid'], 'TestingLinkSourceSno'),
                                           (targets[0]['uuid'], 'TestingLinkTargetSno')}
    assert dummy_request._rev_linked_uuids_by_item == {targets[0]['uuid']: {'reverse': [sources[0]['uuid']]}}
    assert res2['rev_link_names'] == {'reverse': [sources[0]['uuid']]}
    assert res2['rev_linked_to_me'] == []

    # test the next target to ensure that the _sid_cache persists between requests
    # sources[1]['uuid'] does not show up because it has status=deleted (no rev_link)
    res3 = dummy_request.embed('/testing-link-targets-sno/', targets[1]['uuid'], '@@index-data', as_user='INDEXER')
    assert {sources[0]['uuid'], targets[0]['uuid'], targets[1]['uuid']} <= set(dummy_request._sid_cache)
    ignored(res3)



import pytest
from urllib3.util import parse_url

from dcicutils.qa_utils import notice_pytest_fixtures
from .postgresql_fixture import (
    SNOVAULT_DB_TEST_DBNAME, SNOVAULT_DB_TEST_HOSTNAME, SNOVAULT_DB_TEST_PORT, SNOVAULT_DB_TEST_USERNAME,
)


def _check_postgresql_url(url):
    assert url.scheme == 'postgresql'
    assert url.hostname == SNOVAULT_DB_TEST_HOSTNAME
    assert url.port == SNOVAULT_DB_TEST_PORT
    assert url.auth == SNOVAULT_DB_TEST_USERNAME  # No password. That goes elsewhere
    assert url.path == '/' + SNOVAULT_DB_TEST_DBNAME
    assert url.query.startswith("host=")  # The next character is either / or %2F (hopefully the latter)


def test_postgresql_server(postgresql_server, request):
    notice_pytest_fixtures(request)
    url = parse_url(postgresql_server)
    _check_postgresql_url(url)


def test_engine_url(engine_url, request):
    notice_pytest_fixtures(request)
    url = parse_url(engine_url)
    _check_postgresql_url(url)



import copy
import pytest

# from dcicutils.qa_utils import notice_pytest_fixtures
from ..interfaces import TYPES
from ..util import (
    build_embedded_model,
    build_default_embeds,
    crawl_schema,
    expand_embedded_list,
    find_collection_subtypes,
    find_default_embeds_for_schema,
)

# This is handled in pytest.ini and should not be redundantly here. -kmp 4-Jul-2020
# from .toolfixtures import registry
# notice_pytest_fixtures(registry)


def test_find_collection_subtypes(app):
    test_item_type = 'AbstractItemTest'
    expected_types = ['abstract_item_test_second_sub_item', 'abstract_item_test_sub_item']
    res = find_collection_subtypes(app.registry, test_item_type)
    assert sorted(res) == expected_types


def test_build_default_embeds():
    embeds_to_add = ['obj1.obj2', 'obj1', 'obj1.obj3.*']
    processed_embeds = {'obj1.*'}
    final_embeds = build_default_embeds(embeds_to_add, processed_embeds)
    expected_embeds = [
        'obj1.obj2.display_title',
        'obj1.obj2.@id',
        'obj1.obj2.@type',
        'obj1.obj2.uuid',
        'obj1.obj2.principals_allowed.*',
        'obj1.obj2.status',
        'obj1.*',
        'obj1.obj3.*'
    ]
    assert(set(final_embeds) == set(expected_embeds))


def test_find_default_embeds_and_expand_emb_list(registry):
    # use EmbeddingTest as test case
    # 'attachment' -> linkTo TestingDownload
    type_info = registry[TYPES].by_item_type['embedding_test']
    schema_props = type_info.schema.get('properties')
    default_embeds = find_default_embeds_for_schema('', schema_props)
    expected_embeds = ['attachment', 'principals_allowed.*']
    assert(set(default_embeds) == set(expected_embeds))

    # get expansions from 'attachment'
    dummy_emb_list = [
        emb + '.*'
        if not emb.endswith('*') else emb
        for emb in expected_embeds
    ]
    embs_to_add, _ = expand_embedded_list('EmbeddingTest', registry[TYPES], dummy_emb_list, schema_props, set())
    expected_to_add = [
        'attachment',
        'attachment.attachment.*',
        'attachment.attachment2.*',
        'attachment.principals_allowed.*'
    ]
    assert(set(embs_to_add) == set(expected_to_add))

    # add default embeds for all items 'attachment'
    test_embed = ['attachment.attachment.*']
    embs_to_add2, _ = expand_embedded_list('EmbeddingTest', registry[TYPES], test_embed, schema_props, set())
    expected_to_add2 = ['attachment']
    assert(set(embs_to_add2) == set(expected_to_add2))
    # lastly check the built embeds
    expected_built = [
        'attachment.display_title',
        'attachment.@type',
        'attachment.uuid',
        'attachment.@id',
        'attachment.principals_allowed.*',
        'attachment.status'
    ]
    assert set(expected_built) == set(build_default_embeds(embs_to_add2, set()))


def test_crawl_schema(registry):
    field_path = 'attachment.@type'
    embedding_schema = registry[TYPES].by_item_type['embedding_test'].schema
    res = crawl_schema(registry[TYPES], field_path, embedding_schema)
    assert isinstance(res, dict)
    assert res['type'] == 'array'

    # test some bad cases.
    with pytest.raises(Exception) as exec_info:
        crawl_schema(registry[TYPES], field_path, 'not_a_schema')
    # different error, since it attempts to find the file locally
    assert 'Invalid starting schema' in str(exec_info)

    field_path2 = 'attachment.@id.title'
    with pytest.raises(Exception) as exec_info2:
        crawl_schema(registry[TYPES], field_path2, embedding_schema)
    # different error, since it attempts to find the file locally
    assert 'Non-dictionary schema' in str(exec_info2)

    field_path3 = 'attachment.@types.blah'
    with pytest.raises(Exception) as exec_info3:
        crawl_schema(registry[TYPES], field_path3, embedding_schema)
    # different error, since it attempts to find the file locally
    assert 'Field not found' in str(exec_info3)

    # screw with the schema to create an invalid linkTo
    embedding_schema = registry[TYPES].by_item_type['embedding_test'].schema
    schema_copy = copy.deepcopy(embedding_schema)
    schema_copy['properties']['attachment']['linkTo'] = 'NotAnItem'
    with pytest.raises(Exception) as exec_info4:
        crawl_schema(registry[TYPES], field_path, schema_copy)
    # different error, since it attempts to find the file locally
    assert 'Invalid linkTo' in str(exec_info4)


def test_build_embedded_model():
    res1 = build_embedded_model([
        "link_to_nested_objects.display_title",
        "link_to_nested_object.uuid",
        "link_to_nested_object.@id",
        "link_to_nested_object.display_title",
        "link_to_nested_objects.principals_allowed.*",
        "link_to_nested_objects.associates.y",
        "principals_allowed.*",
        "link_to_nested_objects.uuid",
        "link_to_nested_object.principals_allowed.*",
        "link_to_nested_objects.@type",
        "link_to_nested_objects.@id",
        "link_to_nested_object.associates.x",
        "link_to_nested_object.@type"
    ])
    assert res1 == {
        "fields_to_use": ["*"],
        "link_to_nested_objects": {
            "fields_to_use": ["display_title", "uuid", "@type", "@id"],
            "principals_allowed": {
                "fields_to_use": ["*"]},
            "associates": {
                "fields_to_use": ["y"]
            }
        },
        "link_to_nested_object": {
            "fields_to_use": ["uuid", "@id", "display_title", "@type"],
            "principals_allowed": {
                "fields_to_use": ["*"]
            },
            "associates": {
                "fields_to_use": ["x"]
            }
        },
        "principals_allowed": {
            "fields_to_use": ["*"]
        }
    }
    res2 = build_embedded_model([
        "link_to_nested_objects.display_title",
        "link_to_nested_object.uuid",
        "link_to_nested_object.@id",
        "link_to_nested_object.display_title",
        "link_to_nested_objects.principals_allowed.*",
        "link_to_nested_objects.associates.x",
        "link_to_nested_objects.associates.y",
        "principals_allowed.*",
        "link_to_nested_objects.uuid",
        "link_to_nested_object.principals_allowed.*",
        "link_to_nested_objects.@type",
        "link_to_nested_objects.@id",
        "link_to_nested_object.associates.y",
        "link_to_nested_object.associates.x",
        "link_to_nested_object.@type"
    ])
    assert res2 == {
        "fields_to_use": ["*"],
        "link_to_nested_objects": {
            "fields_to_use": ["display_title", "uuid", "@type", "@id"],
            "principals_allowed": {
                "fields_to_use": ["*"]},
            "associates": {
                "fields_to_use": ["x", "y"]
            }
        },
        "link_to_nested_object": {
            "fields_to_use": ["uuid", "@id", "display_title", "@type"],
            "principals_allowed": {
                "fields_to_use": ["*"]
            },
            "associates": {
                "fields_to_use": ["y", "x"]
            }
        },
        "principals_allowed": {
            "fields_to_use": ["*"]
        }
    }



import copy
from pyramid.security import (
    # ALL_PERMISSIONS,
    Allow,
    # Authenticated,
    Deny,
    # DENY_ALL,
    Everyone,
    principals_allowed_by_permission,
)
from pyramid.traversal import find_root, traverse
from pyramid.view import view_config
from sqlalchemy import inspect
from transaction.interfaces import TransientError
from ..resources import (
    AbstractCollection as BaseAbstractCollection,
    Collection as BaseCollection,
    Item as BaseItem,
)
from ..calculated import calculated_property
from ..config import collection, abstract_collection
from ..schema_utils import load_schema
from ..attachment import ItemWithAttachment
from ..interfaces import CONNECTION
# from .root import TestRoot


def includeme(config):
    config.scan(__name__)


# Item acls

def append_acls(acls1: list, acls2: list) -> list:
    # PyCharm gets overly aggressive about type-checking lists of tuples when they are added,
    # fearing every detail of every internal element has to match. This asserts that only list-ness
    # is being relied upon.
    return acls1 + acls2


ONLY_ADMIN_VIEW = [
    (Allow, 'group.admin', ['view', 'edit']),
    (Allow, 'group.read-only-admin', ['view']),
    (Allow, 'remoteuser.INDEXER', ['view']),
    (Allow, 'remoteuser.EMBED', ['view']),
    (Allow, Everyone, ['view', 'edit']),
]

ALLOW_EVERYONE_VIEW = append_acls(
    [
        (Allow, Everyone, ['view', 'list']),
    ],
    ONLY_ADMIN_VIEW)

ALLOW_VIEWING_GROUP_VIEW = append_acls(
    [
        (Allow, 'role.viewing_group_member', 'view'),
    ],
    ONLY_ADMIN_VIEW)

ALLOW_LAB_SUBMITTER_EDIT = append_acls(
    [
        (Allow, 'role.viewing_group_member', 'view'),
        (Allow, 'role.lab_submitter', 'edit'),
    ],
    ONLY_ADMIN_VIEW)

ALLOW_CURRENT_AND_SUBMITTER_EDIT = append_acls(
    [
        (Allow, Everyone, 'view'),
        (Allow, 'role.lab_submitter', 'edit'),
    ],
    ONLY_ADMIN_VIEW)

ALLOW_CURRENT = append_acls(
    [
        (Allow, Everyone, 'view'),
    ],
    ONLY_ADMIN_VIEW)

DELETED = append_acls(
    [
        (Deny, Everyone, 'visible_for_edit'),
    ],
    ONLY_ADMIN_VIEW)


# Collection acls

ALLOW_SUBMITTER_ADD = [
    (Allow, Everyone, ['add']),
]


@view_config(name='testing-user', request_method='GET')
def user(request):
    return {
        'authenticated_userid': request.authenticated_userid,
        'effective_principals': request.effective_principals,
    }


@view_config(name='testing-allowed', request_method='GET')
def allowed(context, request):
    permission = request.params.get('permission', 'view')
    return {
        'has_permission': bool(request.has_permission(permission, context)),
        'principals_allowed_by_permission': principals_allowed_by_permission(context, permission),
    }


def paths_filtered_by_status(request, paths, exclude=('deleted', 'replaced'), include=None):
    """
    This function has been deprecated in Fourfront, but is still used by
    access_keys calc property in types/user.py (only for snowflakes)
    filter out status that shouldn't be visible.
    Also convert path to str as functions like rev_links return uuids
    """
    if include is not None:
        return [
            path for path in paths
            if traverse(request.root, str(path))['context'].__json__(request).get('status') in include
        ]
    else:
        return [
            path for path in paths
            if traverse(request.root, str(path))['context'].__json__(request).get('status') not in exclude
        ]


class AbstractCollection(BaseAbstractCollection):
    def get(self, name, default=None):
        resource = super(AbstractCollection, self).get(name, None)
        if resource is not None:
            return resource
        if ':' in name:
            resource = self.connection.get_by_unique_key('alias', name)
            if resource is not None:
                if not self._allow_contained(resource):
                    return default
                return resource
        return default


class Collection(BaseCollection):
    def __init__(self, *args, **kw):
        super(Collection, self).__init__(*args, **kw)
        if hasattr(self, '__acl__'):
            return
        # XXX collections should be setup after all types are registered.
        # Don't access type_info.schema here as that precaches calculated schema too early.
        self.__acl__ = (ALLOW_SUBMITTER_ADD + ALLOW_EVERYONE_VIEW)


@abstract_collection(
    name='items',
    properties={
        'title': "Item Listing",
        'description': 'Abstract collection of all Items.',
    })
class Item(BaseItem):
    item_type = 'item'
    AbstractCollection = AbstractCollection
    Collection = Collection
    default_diff = []
    STATUS_ACL = {
        # standard_status
        'released': ALLOW_CURRENT,
        'deleted': DELETED,
        'replaced': DELETED,

        # shared_status
        'current': ALLOW_CURRENT,
        'disabled': ONLY_ADMIN_VIEW,

        # file
        'obsolete': ONLY_ADMIN_VIEW,

        # "sets"
        'release ready': ALLOW_VIEWING_GROUP_VIEW,
        'revoked': ALLOW_CURRENT,
        'in review': ALLOW_CURRENT_AND_SUBMITTER_EDIT,

        # publication
        'published': ALLOW_CURRENT,

        # pipeline
        'active': ALLOW_CURRENT,
        'archived': ALLOW_CURRENT,
    }
    filtered_rev_statuses = ('deleted', 'replaced')

    @property
    def __name__(self):
        if self.name_key is None:
            return self.uuid
        properties = self.upgrade_properties()
        if properties.get('status') == 'replaced':
            return self.uuid
        return properties.get(self.name_key, None) or self.uuid

    def __acl__(self):
        # Don't finalize to avoid validation here.
        properties = self.upgrade_properties().copy()
        status = properties.get('status')
        if status is None:
            return [(Allow, Everyone, ['list', 'add', 'view', 'edit', 'add_unvalidated', 'index',
                                       'storage', 'import_items', 'search'])]
        return self.STATUS_ACL.get(status, ALLOW_LAB_SUBMITTER_EDIT)

    def __ac_local_roles__(self):
        roles = {}
        properties = self.upgrade_properties().copy()
        if 'lab' in properties:
            lab_submitters = 'submits_for.%s' % properties['lab']
            roles[lab_submitters] = 'role.lab_submitter'
        if 'award' in properties:
            # TODO: This will fail. There is no function _award_viewing_group anywhere in snovault.
            #       Maybe we should create a hook that Fourfront but not CGAP can set that knows about awards.
            #       Probably 'lab' above has the same issues. -kmp 4-Jul-2020
            viewing_group = _award_viewing_group(properties['award'], find_root(self))
            if viewing_group is not None:
                viewing_group_members = 'viewing_group.%s' % viewing_group
                roles[viewing_group_members] = 'role.viewing_group_member'
        return roles

    def unique_keys(self, properties):
        keys = super(Item, self).unique_keys(properties)
        if 'accession' not in self.schema['properties']:
            return keys
        keys.setdefault('accession', []).extend(properties.get('alternate_accessions', []))
        if properties.get('status') != 'replaced' and 'accession' in properties:
            keys['accession'].append(properties['accession'])
        return keys

    @calculated_property(schema={
        "title": "Display Title",
        "description": "A calculated title for every object in 4DN",
        "type": "string"
    },)
    def display_title(self):
        """create a display_title field."""
        # Unused
        # display_title = ""
        look_for = [
            "title",
            "name",
            "location_description",
            "accession",
        ]
        for field in look_for:
            # special case for user: concatenate first and last names
            display_title = self.properties.get(field, None)
            if display_title:
                return display_title
        # if none of the existing terms are available, use @type + date_created
        try:
            # TODO: PyCharm thinks self.__class__.__name__ will return 'property' not 'string'.
            #       That's probably a bug in PyCharm I should report. -kmp 4-Jul-2020
            type_date = self.__class__.__name__ + " from " + self.properties.get("date_created", None)[:10]
            return type_date
        # last resort, use uuid
        except Exception:
            return self.properties.get('uuid', None)


@calculated_property(context=Item.Collection, category='action')
def add(context, request):
    if request.has_permission('add'):
        return {
            'name': 'add',
            'title': 'Add',
            'profile': '/profiles/{ti.name}.json'.format(ti=context.type_info),
            'href': '{item_uri}#!add'.format(item_uri=request.resource_path(context)),
            }


@calculated_property(context=Item, category='action')
def edit(context, request):
    if request.has_permission('edit'):
        return {
            'name': 'edit',
            'title': 'Edit',
            'profile': '/profiles/{ti.name}.json'.format(ti=context.type_info),
            'href': '{item_uri}#!edit'.format(item_uri=request.resource_path(context)),
        }


@calculated_property(context=Item, category='action')
def edit_json(context, request):
    if request.has_permission('edit'):
        return {
            'name': 'edit-json',
            'title': 'Edit JSON',
            'profile': '/profiles/{ti.name}.json'.format(ti=context.type_info),
            'href': '{item_uri}#!edit-json'.format(item_uri=request.resource_path(context)),
        }


@abstract_collection(
    name='abstractItemTests',
    unique_key='accession',
    properties={
        'title': "AbstractItemTests",
        'description': "Abstract Item that is inherited for testing",
    })
class AbstractItemTest(Item):
    item_type = 'AbstractItemTest'
    base_types = ['AbstractItemTest'] + Item.base_types
    name_key = 'accession'


@collection(
    name='abstract-item-test-sub-items',
    unique_key='accession',
    properties={
        'title': "AbstractItemTestSubItems",
        'description': "Item based off of AbstractItemTest"
    })
class AbstractItemTestSubItem(AbstractItemTest):
    item_type = 'abstract_item_test_sub_item'
    schema = load_schema('snovault:test_schemas/AbstractItemTestSubItem.json')


@collection(
    name='abstract-item-test-second-sub-items',
    unique_key='accession',
    properties={
        'title': 'AbstractItemTestSecondSubItems',
        'description': "Second item based off of AbstractItemTest"
    })
class AbstractItemTestSecondSubItem(AbstractItemTest):
    item_type = 'abstract_item_test_second_sub_item'
    schema = load_schema('snovault:test_schemas/AbstractItemTestSecondSubItem.json')


@collection(
    name='embedding-tests',
    unique_key='accession',
    properties={
        'title': 'EmbeddingTests',
        'description': 'Listing of EmbeddingTests'
    })
class EmbeddingTest(Item):
    item_type = 'embedding_test'
    schema = load_schema('snovault:test_schemas/EmbeddingTest.json')
    name_key = 'accession'

    # use TestingDownload to test
    embedded_list = [
        'attachment.*'
    ]

# Formerly b58bc82f-249e-418f-bbcd-8a80af2e58d3
NESTED_OBJECT_LINK_TARGET_GUID_1 = 'f738e192-85f4-4886-bdc4-e099a2e2102a'
NESTED_OBJECT_LINK_TARGET_GUID_2 = 'c48dfba9-ad62-4b32-ad29-a4b6ca47e5d4'

# Formerly 100a0bb8-2974-446b-a5de-6937aa313be4
NESTED_EMBEDDING_CONTAINER_GUID = "6d3e9e27-cf87-4103-aa36-9f481c9d9a66"

NESTED_OBJECT_LINK_TARGET_GUIDS = [  # These IDs are defined in test_views.py so this is a low-tech revlink
    NESTED_OBJECT_LINK_TARGET_GUID_1,
    NESTED_OBJECT_LINK_TARGET_GUID_2,
]


@collection(
    name='nested-embedding-container',
    unique_key='accession',
    properties={
        'title': 'NestedEmbeddingContainer',
        'description': 'Test of ...'
    })
class NestedEmbeddingContainer(Item):
    item_type = 'nested_embedding_container'
    schema = load_schema('snovault:test_schemas/NestedEmbeddingContainer.json')
    name_key = 'accession'

    # use TestingDownload to test
    embedded_list = [
        'link_to_nested_object.associates.x',
        'link_to_nested_object.associates.y',
        'link_to_nested_objects.associates.x',
        'link_to_nested_objects.associates.y',
        'nested_calculated_property.associates.x',
        'nested_calculated_property.associates.y',
    ]

    @calculated_property(schema={
            "title": "Nested Calculated property",
            "description": "something calculated",
            "type": "array",
            "items": {
                "title": "Nested Calculated Property",
                "type": ["string", "object"],
                "linkTo": "NestedObjectLinkTarget"
            }
        })
    def nested_calculated_property(self):
        return copy.copy(NESTED_OBJECT_LINK_TARGET_GUIDS)


@collection(
    name='nested-object-link-target',
    unique_key='accession',
    properties={
        'title': 'NestedObjectLinkTarget',
        'description': '...'
    })
class NestedObjectLinkTarget(Item):
    item_type = 'nested_object_link_target'
    schema = load_schema('snovault:test_schemas/NestedObjectLinkTarget.json')
    name_key = 'accession'


@collection(
    'testing-downloads',
    properties={
        'title': 'Test download collection',
        'description': 'Testing. Testing. 1, 2, 3.',
    },
)
class TestingDownload(ItemWithAttachment):
    item_type = 'testing_download'
    schema = load_schema('snovault:test_schemas/TestingDownload.json')


@collection('testing-link-sources-sno', unique_key='testing_link_sources-sno:name')
class TestingLinkSourceSno(Item):
    item_type = 'testing_link_source_sno'
    schema = load_schema('snovault:test_schemas/TestingLinkSourceSno.json')
    embedded_list = ['target_es.status', 'target.status']


@collection('testing-link-aggregates-sno')
class TestingLinkAggregateSno(Item):
    item_type = 'testing_link_aggregate_sno'
    schema = load_schema('snovault:test_schemas/TestingLinkAggregateSno.json')
    aggregated_items = {
        "targets": ['target.uuid', 'test_description']
    }


@collection('testing-link-targets-sno', unique_key='testing_link_target_sno:name')
class TestingLinkTargetSno(Item):
    item_type = 'testing_link_target_sno'
    name_key = 'name'
    schema = load_schema('snovault:test_schemas/TestingLinkTargetSno.json')
    rev = {
        'reverse': ('TestingLinkSourceSno', 'target'),
    }
    filtered_rev_statuses = ('deleted', 'replaced')
    embedded_list = [
        'reverse.name',
    ]

    def rev_link_atids(self, request, rev_name):
        conn = request.registry[CONNECTION]
        return [request.resource_path(conn[uuid]) for uuid in
                self.get_filtered_rev_links(request, rev_name)]

    @calculated_property(schema={
        "title": "Sources",
        "type": "array",
        "items": {
            "type": ['string', 'object'],
            "linkTo": "TestingLinkSourceSno",
        },
    })
    def reverse(self, request):
        return self.rev_link_atids(request, "reverse")


# Renamed from TestingPostPutPatch to TestingPostPutPatchSno so that indices
# would not coincide with Fourfront tests, which also use that index name
@collection(
    'testing-post-put-patch-sno',
    acl=[
        (Allow, 'group.submitter', ['add', 'edit', 'view']),
    ],
)
class TestingPostPutPatchSno(Item):
    item_type = 'testing_post_put_patch_sno'
    embedded_list = ['protected_link.*']
    schema = load_schema('snovault:test_schemas/TestingPostPutPatchSno.json')


@collection('testing-server-defaults')
class TestingServerDefault(Item):
    item_type = 'testing_server_default'
    schema = load_schema('snovault:test_schemas/TestingServerDefault.json')


@collection('testing-dependencies')
class TestingDependencies(Item):
    item_type = 'testing_dependencies'
    schema = load_schema('snovault:test_schemas/TestingDependencies.json')


@view_config(name='testing-render-error', request_method='GET')
def testing_render_error(request):
    return {
        '@type': ['TestingRenderError', 'Item'],
        '@id': request.path,
        'title': 'Item triggering a render error',
    }


@view_config(context=TestingPostPutPatchSno, name='testing-retry')
def testing_retry(context, request):

    model = context.model
    request.environ['_attempt'] = request.environ.get('_attempt', 0) + 1

    if request.environ['_attempt'] == 1:
        raise TransientError()

    return {
        'attempt': request.environ['_attempt'],
        'detached': inspect(model).detached,
    }


# properties_datastore sets makes this collection stored in ES
@collection('testing-link-targets-elastic-search',
            unique_key='testing_link_target_elastic_search:name',
            properties_datastore='elasticsearch')
class TestingLinkTargetElasticSearch(Item):
    """
    Like TestingLinkTargetSno, but leverages ElasticSearch storage exclusively.
    Includes a linkTo and a rev_link to test multiple behaviors.
    """
    item_type = 'testing_link_target_elastic_search'
    name_key = 'name'
    schema = load_schema('snovault:test_schemas/TestingLinkTargetElasticSearch.json')
    rev = {
        'reverse_es': ('TestingLinkSourceSno', 'target_es'),
    }
    filtered_rev_statuses = ('deleted', 'replaced')
    aggregated_items = {
        "ppp": ['simple1', 'uuid']
    }
    embedded_list = [
        'reverse_es.name',
        'ppp.simple1'
    ]

    def rev_link_atids(self, request, rev_name):
        conn = request.registry[CONNECTION]
        return [request.resource_path(conn[uuid]) for uuid in
                self.get_filtered_rev_links(request, rev_name)]

    @calculated_property(schema={
        "title": "Sources",
        "type": "array",
        "items": {
            "type": ['string', 'object'],
            "linkTo": "TestingLinkSourceSno",
        },
    })
    def reverse_es(self, request):
        return self.rev_link_atids(request, "reverse_es")


@collection('testing-calculated-properties',
            unique_key='testing_calculated_properties:name')
class TestingCalculatedProperties(Item):
    """ An item type that has calculated properties on it meant for testing. """
    item_type = 'testing_calculated_properties'
    name_key = 'name'
    schema = load_schema('snovault:test_schemas/TestingCalculatedProperties.json')

    @calculated_property(schema={
        "title": "combination",
        "type": "object"
    })
    def combination(self, name, foo, bar):
        return {
            'name': name,
            'foo': foo,
            'bar': bar
        }

    @calculated_property(schema={  # THIS is the schema that will be "seen"
        "title": "nested",
        "type": "object",
        "sub-embedded": True,  # REQUIRED TO INDICATE
        "properties": {
            "key": {
                "type": "string"
            },
            "value": {
                "type": "string"
            },
            "keyvalue": {
                "type": "string"
            }
        }
    })
    def nested(self, nested):  # nested is the calculated property path that will update and the input
        """ Implements sub-embedded-object calculated properties.

            When merged into properties looks like this:
            {
                'nested' : {
                    'keyvalue': val
                }
            }
        """
        # return a dictionary with all sub-embedded key, value pairs on this sub-embedded path
        return {'keyvalue': nested['key'] + nested['value']}

    @calculated_property(schema={  # IN ORDER TO GET CORRECT MAPPINGS, YOU MUST SPECIFY THE ENTIRE SCHEMA
        "title": "nested2",
        "type": "array",
        "sub-embedded": True,
        "items": {
            "type": "object",
            "properties": {
                "key": {
                    "type": "string"
                },
                "value": {
                    "type": "string"
                },
                "keyvalue": {
                    "type": "string"
                }
            }
        }
    })
    def nested2(self, nested2):
        """ Implements sub-embedded object calculated property on array (of objects) type field

            When merged into properties looks like this:
            {
                'nested2': [
                    {
                        keyvalue: val
                    },

                    {
                        keyvalue: val
                    }
                ]
            }
        """
        # return an ARRAY of dictionaries
        result = []
        for entry in nested2:
            result.append({
                'keyvalue': entry['key'] + entry['value']
            })
        return result


@collection('testing-mixins', unique_key='testing_mixins:name')
class TestingMixins(Item):
    item_type = 'testing_mixins'
    name_key = 'name'
    schema = load_schema('snovault:test_schemas/TestingMixins.json')


@collection('testing-nested-enabled', unique_key='testing_nested_enabled:name')
class TestingNestedEnabled(Item):
    """ Type intended to test enabling nested mappings per-field. """
    item_type = 'testing_nested_enabled'
    name_key = 'name'
    schema = load_schema('snovault:test_schemas/TestingNestedEnabled.json')

    @calculated_property(schema={
        "title": "enabled_array_of_objects_in_calc_prop",
        "description": "Tests mapping calculated properties with enable_nested works correctly",
        "type": "array",
        "items": {
            "type": "object",
            "enable_nested": True,
            "properties": {
                "string_field": {
                    "type": "string"
                },
                "numerical_field": {
                    "type": "integer"
                }
            }
        }
    })
    def enabled_array_of_objects_in_calc_prop(self):
        """ This one will get mapped with nested """
        return [{
            'string_field': 'hello',
            'numerical_field': 0
        }]

    @calculated_property(schema={
        "title": "array_of_objects_in_calc_prop",
        "description": "Tests mapping calculated properties with disable_nested works correctly",
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "string_field": {
                    "type": "string"
                },
                "numerical_field": {
                    "type": "integer"
                }
            }
        }
    })
    def disabled_array_of_objects_in_calc_prop(self):
        """ This one will not get mapped with nested since it was not explicitly enabled """
        return [{
            'string_field': 'world',
            'numerical_field': 100
        }]


@collection(name='testing-individual-sno', unique_key='testing_individual_sno:full_name')
class TestingIndividualSno(Item):
    """ Individual integrated testing type - a biosample is produced by an individual. """
    item_type = 'testing_individual_sno'
    name_key = 'full_name'
    schema = load_schema('snovault:test_schemas/TestingIndividualSno.json')

    @calculated_property(schema={
        "title": "Last Name",
        "type": "string",
    })
    def last_name(self, full_name):
        return full_name.split()[1]


@collection(name='testing-biosample-sno', unique_key='testing_biosample_sno:identifier')
class TestingBiosampleSno(Item):
    """ Biosample integrated testing type. """
    item_type = 'testing_biosample_sno'
    name_key = 'identifier'
    schema = load_schema('snovault:test_schemas/TestingBiosampleSno.json')
    embedded_list = [
        'contributor.specimen'
    ]


@collection(name='testing-biosource-sno', unique_key='testing_biosource_sno:identifier')
class TestingBiosourceSno(Item):
    """ Biosource integrated testing type.

        This item contains an array of linkTo Biosamples. Links to a contributor of the samples as well.
    """
    item_type = 'testing_biosource_sno'
    name_key = 'identifier'
    schema = load_schema('snovault:test_schemas/TestingBiosourceSno.json')
    embedded_list = [  # selective embed at this item
        'samples.identifier',
        'samples.quality',
        'sample_objects.associated_sample.alias',  # if a sample is included here, embed its alias
        'contributor.full_name',
        'contributor.last_name'  # calc prop, dependent on above embed
    ]
    default_diff = [  # the counter is updated
        'counter'
    ]

    def _update(self, properties, sheets=None):
        """ Updates the counter in addition """
        if 'counter' in properties:
            properties['counter'] += 1
        else:
            properties['counter'] = 1
        super(TestingBiosourceSno, self)._update(properties, sheets)


@collection(name='testing-biogroup-sno', unique_key='testing_biogroup_sno:name')
class TestingBiogroupSno(Item):
    """ Biogroup integrated testing type.

        Biogroup consists of an array of Biosource objects that consist of an array of Biosample objects.
    """
    item_type = 'testing_biogroup_sno'
    name_key = 'name'
    schema = load_schema('snovault:test_schemas/TestingBiogroupSno.json')
    embedded_list = [
        'sources.counter',  # get the counter
        'sources.samples.*',  # embed everything at top level
        'sources.contributor.*'
    ]



from __future__ import unicode_literals

import sys

from contextlib import contextmanager
from timeit import default_timer as timer

from pyramid.settings import asbool
from pyramid.traversal import resource_path
from pyramid.view import view_config
from pyramid.settings import asbool
from timeit import default_timer as timer
from contextlib import contextmanager

from .resources import Item
from .interfaces import STORAGE
from .elasticsearch.indexer_utils import find_uuids_for_indexing
from .embed import make_subrequest
from .interfaces import STORAGE
from .resources import Item
from .util import debug_log
from .validation import ValidationFailure


def includeme(config):
    config.add_route('indexing-info', '/indexing-info')
    config.add_route('max-sid', '/max-sid')
    config.scan(__name__)


@contextmanager
def indexing_timer(timer_dict, time_key):
    """
    Simple contextmanager to time components of index-data

    Args:
        timer_dict (dict): dictionary to add timing results to
        time_key (str): key to use for the given timing result
    """
    start = timer()
    yield
    timer_dict[time_key] = timer() - start


def join_linked_uuids_sids(request, uuid_type_pairs):
    """
    Simply iterate through the uuid_type_pairs and return an array of dicts containing
    uuid, sid (from request._sid_cache) and the item type.

    Args:
        request: current Request object
        uuid_type_pairs: list of 2-tuples (uuid, item_type)

    Returns:
        A list of dicts containing uuid, up-to-date db sid and item type
    """
    return [{'uuid': uuid,
             'sid': request._sid_cache[uuid],
             'item_type': item_type} for uuid, item_type in uuid_type_pairs]


def get_rev_linked_items(request, uuid):
    """
    Iterate through request._rev_linked_uuids_by_item, which is populated
    during the embedding traversal process, to find the items that are reverse
    linked to the given uuid

    Args:
        request: current Request object
        uuid (str): uuid of the object in question

    Returns:
        A set of string uuids
    """
    # find uuids traversed that rev link to this item
    rev_linked_to_me = set()
    for rev_id, rev_names in request._rev_linked_uuids_by_item.items():
        if any([uuid in rev_names[name] for name in rev_names]):
            rev_linked_to_me.add(rev_id)
            continue
    return rev_linked_to_me


@view_config(context=Item, name='index-data', permission='index', request_method='GET')
@debug_log
def item_index_data(context, request):
    """
    Very important view which is used to calculate all the data indexed in ES
    for the given item. If an int sid is provided as a request parameter,
    will raise an sid exception if the maximum sid value on current_propsheets
    table is less than the given value.

    Computationally intensive. Calculates the object and embedded views
    for the given item, using ES results where possible for speed. Also handles
    calculation of aggregated-items and validation-errors for the item.
    Leverages a number of attrs on the request to get needed information

    Args:
        context: current Item
        request: current Request

    Returns:
        A dict document representing the full data to index for the given item
    """
    indexing_stats = {}  # hold timing details for this view

    uuid = str(context.uuid)

    # upgrade_properties calls necessary upgraders based on schema_version
    with indexing_timer(indexing_stats, 'upgrade_properties'):
        properties = context.upgrade_properties()

    # ES versions 2 and up don't allow dots in links. Update these to use ~s
    new_links = {}
    for key, val in context.links(properties).items():
        new_links['~'.join(key.split('.'))] = val
    links = new_links

    principals_allowed = context.principals_allowed()
    path = resource_path(context)
    paths = {path}
    collection = context.collection

    with indexing_timer(indexing_stats, 'unique_keys'):
        unique_keys = context.unique_keys(properties)
        if collection.unique_key in unique_keys:
            paths.update(
                resource_path(collection, key)
                for key in unique_keys[collection.unique_key])

    with indexing_timer(indexing_stats, 'paths'):
        for base in (collection, request.root):
            for key_name in ('accession', 'alias'):
                if key_name not in unique_keys:
                    continue
                paths.add(resource_path(base, uuid))
                paths.update(
                    resource_path(base, key)
                    for key in unique_keys[key_name])

    path = path + '/'
    # setting _indexing_view enables the embed_cache and cause population of
    # request._linked_uuids and request._rev_linked_uuids_by_item
    request._indexing_view = True

    # run the object view first
    request._linked_uuids = set()
    with indexing_timer(indexing_stats, 'object_view'):
        object_view = request.invoke_view(path, '@@object')
    linked_uuids_object = request._linked_uuids.copy()
    rev_link_names = request._rev_linked_uuids_by_item.get(uuid, {}).copy()

    # reset these properties, then run embedded view
    request._linked_uuids = set()
    request._rev_linked_uuids_by_item = {}
    request._aggregate_for['uuid'] = uuid
    request._aggregated_items = {
        agg: {'_fields': context.aggregated_items[agg], 'items': []} for agg in context.aggregated_items
    }
    # since request._indexing_view is set to True in indexer.py,
    # all embeds (including subrequests) below will use the embed cache
    with indexing_timer(indexing_stats, 'embedded_view'):
        embedded_view = request.invoke_view(path, '@@embedded', index_uuid=uuid)

    # this is built since the embedded view is built on "item_with_links", see resources.py
    linked_uuids_embedded = request._linked_uuids.copy()

    # find uuids traversed that rev link to this item
    with indexing_timer(indexing_stats, 'rev_links'):
        rev_linked_to_me = get_rev_linked_items(request, uuid)

    # calculated aggregated items
    with indexing_timer(indexing_stats, 'aggregated_items'):
        aggregated_items = {agg: res['items'] for agg, res in
                            request._aggregated_items.items()}

    # run validators for the item by PATCHing with check_only=True
    # json_body provided is the upgraded properties of the item
    with indexing_timer(indexing_stats, 'validation'):
        validate_path = path + '?check_only=true'
        validate_req = make_subrequest(request, validate_path,
                                       json_body=properties)
        try:
            request.invoke_subrequest(validate_req)
        except ValidationFailure:
            # TODO: This should probably be logged. -kmp 22-Oct-2020
            pass

    document = {
        'aggregated_items': aggregated_items,
        'embedded': embedded_view,
        'indexing_stats': indexing_stats,
        'item_type': context.type_info.item_type,
        'linked_uuids_embedded': join_linked_uuids_sids(request, linked_uuids_embedded),
        'linked_uuids_object': join_linked_uuids_sids(request, linked_uuids_object),
        'links': links,
        'max_sid': context.max_sid,
        'object': object_view,
        'paths': sorted(paths),
        'principals_allowed': principals_allowed,
        'properties': properties,
        'propsheets': {
            name: context.propsheets[name]
            for name in context.propsheets.keys() if name != ''
        },
        'rev_link_names': rev_link_names,
        'rev_linked_to_me': sorted(rev_linked_to_me),
        'sid': context.sid,
        'unique_keys': unique_keys,
        'uuid': uuid,
        'validation_errors': validate_req.errors
    }

    return document


@view_config(route_name='indexing-info', permission='index', request_method='GET')
@debug_log
def indexing_info(context, request):
    """
    Endpoint to check some indexing-related properties of a given uuid, which
    is provided using the `uuid=` query parameter. This route cannot be defined
    with the context of a specific Item because that will cause the underlying
    request to use a cached view from Elasticsearch and not properly run
    the @@embedded view from the database.

    If you do not want to calculate the embedded object, use `run=False`

    Args:
        request: current Request object

    Returns:
        dict response
    """
    uuid = request.params.get('uuid')
    if not uuid:
        return {'status': 'error', 'title': 'Error', 'message': 'ERROR! Provide a uuid to the query.'}

    db_sid = request.registry[STORAGE].write.get_by_uuid(uuid).sid
    # es_model will be None if the item is not yet indexed
    es_model = request.registry[STORAGE].read.get_by_uuid(uuid)
    es_sid = es_model.sid if es_model is not None else None
    response = {'sid_db': db_sid, 'sid_es': es_sid, 'title': 'Indexing Info for %s' % uuid}
    if asbool(request.params.get('run', True)):
        request._indexing_view = True
        request.datastore = 'database'
        path = '/' + uuid + '/@@index-data'
        index_view = request.invoke_view(path, index_uuid=uuid, as_user='INDEXER')
        response['indexing_stats'] = index_view['indexing_stats']
        # since there is no diff, we cannot compute invalidation scope here.
        es_assc_uuids, _ = find_uuids_for_indexing(request.registry, set([uuid]))
        new_rev_link_uuids = get_rev_linked_items(request, uuid)
        # invalidated: items linking to this in es + newly rev linked items
        response['uuids_invalidated'] = list(es_assc_uuids | new_rev_link_uuids)
        response['description'] = 'Using live results for embedded view of %s. Query with run=False to skip this.' % uuid
    else:
        response['description'] = 'Query with run=True to calculate live information on invalidation and embedding time.'
    response['display_title'] = 'Indexing Info for %s' % uuid
    response['status'] = 'success'
    return response


@view_config(route_name='max-sid', permission='index', request_method='GET')
@debug_log
def max_sid(context, request):
    """
    Very simple endpoint to return the current maximum sid used in postgres.
    Might make more sense to define this view in storage.py, but leave it here
    with the other sid/indexing related code.

    Args:
        request: current Request object

    Returns:
        dict response
    """
    response = {'display_title': 'Current maximum database sid'}
    try:
        max_sid = request.registry[STORAGE].write.get_max_sid()
        response.update({'status': 'success', 'max_sid': max_sid})
    except Exception as exc:
        response.update({'status': 'failure', 'detail': str(exc)})
    return response



import sys
import logging
from dcicutils.es_utils import create_es_client


logger = logging.getLogger(__name__)


def main():
    logging.basicConfig()
    if len(sys.argv) < 3:
        logger.error('Usage: python wipe_test_indices.py <TRAVIS_JOB_ID> <REMOTE_ES>\n')
        exit(1)

    jid, r_es = sys.argv[1], sys.argv[2]
    logger.info('Wiping ES instances on %s with prefix %s\n' % (r_es, jid))
    try:
        client = create_es_client(r_es, use_aws_auth=True)
    except:
        logger.error('Failed to get ES client')
        exit(1)

    try:
        client.indices.delete(index=jid+'*')
        logger.info('Successfully deleted indices with prefix %s' % jid)
    except Exception as exc:
        logger.error('Failed to delete indices with exception: %s\n' % str(exc))
        exit(1)

if __name__ == '__main__':
    main()



import psutil
import time
import pyramid.tweens
from sqlalchemy import event
from sqlalchemy.engine import Engine
from urllib.parse import urlencode
from .util import get_root_request
from uuid import uuid4


def includeme(config):
    config.add_tween('snovault.stats.stats_tween_factory', under=pyramid.tweens.INGRESS)

from structlog import get_logger
log = get_logger()


def requests_timing_hook(prefix='requests'):
    count_key = prefix + '_count'
    time_key = prefix + '_time'

    def response_hook(r, *args, **kwargs):
        request = get_root_request()
        if request is None:
            return

        stats = request._stats
        stats[count_key] = stats.get(count_key, 0) + 1
        # requests response.elapsed is a timedelta
        e = r.elapsed
        duration = (e.days * 86400 + e.seconds) * 1000000 + e.microseconds
        stats[time_key] = stats.get(time_key, 0) + duration

    return response_hook


# See http://www.sqlalchemy.org/trac/wiki/UsageRecipes/Profiling
@event.listens_for(Engine, 'before_cursor_execute')
def before_cursor_execute(
        conn, cursor, statement, parameters, context, executemany):
    context._query_start_time = int(time.time() * 1e6)


@event.listens_for(Engine, 'after_cursor_execute')
def after_cursor_execute(
        conn, cursor, statement, parameters, context, executemany):
    end = int(time.time() * 1e6)

    request = get_root_request()
    if request is None:
        return

    stats = request._stats
    stats['db_count'] = stats.get('db_count', 0) + 1
    duration = end - context._query_start_time
    stats['db_time'] = stats.get('db_time', 0) + duration


# http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/hooks.html#creating-a-tween-factory
def stats_tween_factory(handler, registry):
    process = psutil.Process()

    def stats_tween(request):
        stats = request._stats = {}

        # set telemetry_id as logger context if passed in
        log_keys = {}
        if 'telemetry_id' in request.params:
            log_keys['telemetry_id'] = request.params['telemetry_id']
        if 'log_action' in request.params:
            log_keys['log_action'] = request.params['log_action']
        if request.registry.settings.get('env_name'):
            log_keys['ff_env'] = request.registry.settings.get('env_name')

        log_keys['url_path'] = request.path
        log_keys['url_qs'] = request.query_string
        log_keys['host'] = request.host.split(":")[0]

        log.bind(**log_keys)
        stats.update(log_keys)

        rss_begin = stats['rss_begin'] = process.memory_info().rss
        begin = stats['wsgi_begin'] = int(time.time() * 1e6)
        response = handler(request)
        end = stats['wsgi_end'] = int(time.time() * 1e6)
        rss_end = stats['rss_end'] = process.memory_info().rss
        stats['wsgi_time'] = end - begin
        stats['rss_change'] = rss_end - rss_begin

        environ = request.environ
        if 'mod_wsgi.queue_start' in environ:
            queue_begin = int(environ['mod_wsgi.queue_start'])
            stats['queue_begin'] = queue_begin
            stats['queue_time'] = begin - queue_begin

        xs = response.headers['X-Stats'] = str(urlencode(sorted(stats.items())))
        if getattr(request, '_add_stats_cookie', False):
            response.set_cookie('X-Stats', xs)

        # log all this stuff
        log.bind(**stats).info("Request timings")

        return response

    return stats_tween
